@@ -195,6 +195,15 @@ class PtyProcess(object):
         command = command_with_path
         argv[0] = command
 
+        # [issue #119] To prevent the case where exec fails and the user is
+        # stuck interacting with a python child process instead of whatever
+        # was expected, we implement the solution from
+        # http://stackoverflow.com/a/3703179 to pass the exception to the
+        # parent process
+
+        # [issue #119] 1. Before forking, open a pipe in the parent process.
+        exec_err_pipe_read, exec_err_pipe_write = os.pipe()
+
         if use_native_pty_fork:
             pid, fd = pty.fork()
         else:
@@ -221,9 +230,16 @@ class PtyProcess(object):
                     if err.args[0] not in (errno.EINVAL, errno.ENOTTY):
                         raise
 
-            # Do not allow child to inherit open file descriptors from parent.
+            # [issue #119] 3. The child closes the reading end and sets the
+            # close-on-exec flag for the writing end.
+            os.close(exec_err_pipe_read)
+            fcntl.fcntl(exec_err_pipe_write, fcntl.F_SETFD, fcntl.FD_CLOEXEC)
+
+            # Do not allow child to inherit open file descriptors from parent,
+            # with the exception of the exec_err_pipe_write of the pipe
             max_fd = resource.getrlimit(resource.RLIMIT_NOFILE)[0]
-            os.closerange(3, max_fd)
+            os.closerange(3, exec_err_pipe_write)
+            os.closerange(exec_err_pipe_write+1, max_fd)
 
             if cwd is not None:
                 os.chdir(cwd)
@@ -231,10 +247,18 @@ class PtyProcess(object):
             for func in before_exec:
                 func()
 
-            if env is None:
-                os.execv(command, argv)
-            else:
-                os.execvpe(command, argv, env)
+            try:
+                if env is None:
+                    os.execv(command, argv)
+                else:
+                    os.execvpe(command, argv, env)
+            except OSError as err:
+                # [issue #119] 5. If exec fails, the child writes the error
+                # code back to the parent using the pipe, then exits.
+                errbytes = str(err).encode('utf-8') if PY3 else str(err)
+                os.write(exec_err_pipe_write, struct.pack('i', err.errno) + errbytes)
+                os.close(exec_err_pipe_write)
+                os._exit(os.EX_OSERR)
 
         # Parent
         inst = cls(pid, fd)
@@ -245,7 +269,30 @@ class PtyProcess(object):
             inst.env = env
         if cwd is not None:
             inst.launch_dir = cwd
-        
+
+        # [issue #119] 2. After forking, the parent closes the writing end
+        # of the pipe and reads from the reading end.
+        os.close(exec_err_pipe_write)
+        exec_err_data = os.read(exec_err_pipe_read, 4096)
+        os.close(exec_err_pipe_read)
+
+        # [issue #119] 6. The parent reads eof (a zero-length read) if the
+        # child successfully performed exec, since close-on-exec made
+        # successful exec close the writing end of the pipe. Or, if exec
+        # failed, the parent reads the error code and can proceed
+        # accordingly. Either way, the parent blocks until the child calls
+        # exec.
+        if len(exec_err_data) != 0:
+
+            exec_err_fmt = 'i'
+            exec_err_fmt_size = struct.calcsize(exec_err_fmt)
+            exec_err_errno = struct.unpack(exec_err_fmt, exec_err_data[:exec_err_fmt_size])[0]
+            exec_err_msg = exec_err_data[exec_err_fmt_size:]
+
+            exception = OSError(exec_err_msg.decode('utf-8', 'replace'))
+            exception.errno = exec_err_errno
+            raise exception
+
         try:
             inst.setwinsize(24, 80)
         except IOError as err:
@@ -421,7 +468,10 @@ class PtyProcess(object):
         self.echo = state
 
     def read(self, size=1024):
-        """Read, and convert different platforms' EOF indications to EOFError.
+        """Read and return at most ``size`` bytes from the pty.
+
+        Can block if there is nothing to read. Raises :exc:`EOFError` if the
+        terminal was closed.
         
         Unlike Pexpect's ``read_nonblocking`` method, this doesn't try to deal
         with the vagaries of EOF on platforms that do strange things, like IRIX
@@ -445,6 +495,11 @@ class PtyProcess(object):
         return s
 
     def readline(self):
+        """Read one line from the pseudoterminal, and return it as unicode.
+
+        Can block if there is nothing to read. Raises :exc:`EOFError` if the
+        terminal was closed.
+        """
         try:
             s = self.fileobj.readline()
         except (OSError, IOError) as err:
@@ -461,10 +516,10 @@ class PtyProcess(object):
         return s
 
     def write(self, s):
-        '''Write data to the pseudoterminal.
+        """Write bytes to the pseudoterminal.
         
-        Returns the number of bytes/characters written.
-        '''
+        Returns the number of bytes written.
+        """
         return self.fileobj.write(s)
 
     def sendcontrol(self, char):
@@ -480,7 +535,8 @@ class PtyProcess(object):
         a = ord(char)
         if a >= 97 and a <= 122:
             a = a - ord('a') + 1
-            return self.fileobj.write(_byte(a))
+            byte = _byte(a)
+            return self.fileobj.write(_byte(a)), byte
         d = {'@': 0, '`': 0,
             '[': 27, '{': 27,
             '\\': 28, '|': 28,
@@ -503,13 +559,13 @@ class PtyProcess(object):
         It is the responsibility of the caller to ensure the eof is sent at the
         beginning of a line. '''
 
-        self.fileobj.write(_EOF)
+        return self.fileobj.write(_EOF), _EOF
 
     def sendintr(self):
         '''This sends a SIGINT to the child. It does not require
         the SIGINT to be the first character on a line. '''
 
-        self.fileobj.write(_EOF)
+        return self.fileobj.write(_INTR), _INTR
 
     def eof(self):
         '''This returns True if the EOF exception was ever raised.
@@ -664,34 +720,42 @@ class PtyProcess(object):
         return False
 
     def kill(self, sig):
-        '''This sends the given signal to the child application. In keeping
-        with UNIX tradition it has a misleading name. It does not necessarily
-        kill the child unless you send the right signal. '''
+        """Send the given signal to the child application.
+
+        In keeping with UNIX tradition it has a misleading name. It does not
+        necessarily kill the child unless you send the right signal. See the
+        :mod:`signal` module for constants representing signal numbers.
+        """
 
         # Same as os.kill, but the pid is given for you.
         if self.isalive():
             os.kill(self.pid, sig)
 
     def getwinsize(self):
-        '''This returns the terminal window size of the child tty. The return
-        value is a tuple of (rows, cols). '''
-
+        """Return the window size of the pseudoterminal as a tuple (rows, cols).
+        """
         TIOCGWINSZ = getattr(termios, 'TIOCGWINSZ', 1074295912)
         s = struct.pack('HHHH', 0, 0, 0, 0)
         x = fcntl.ioctl(self.fd, TIOCGWINSZ, s)
         return struct.unpack('HHHH', x)[0:2]
 
     def setwinsize(self, rows, cols):
-        '''This sets the terminal window size of the child tty. This will cause
-        a SIGWINCH signal to be sent to the child. This does not change the
-        physical window size. It changes the size reported to TTY-aware
-        applications like vi or curses -- applications that respond to the
-        SIGWINCH signal. '''
+        """Set the terminal window size of the child tty.
 
+        This will cause a SIGWINCH signal to be sent to the child. This does not
+        change the physical window size. It changes the size reported to
+        TTY-aware applications like vi or curses -- applications that respond to
+        the SIGWINCH signal.
+        """
         return _setwinsize(self.fd, rows, cols)
 
 
 class PtyProcessUnicode(PtyProcess):
+    """Unicode wrapper around a process running in a pseudoterminal.
+
+    This class exposes a similar interface to :class:`PtyProcess`, but its read
+    methods return unicode, and its :meth:`write` accepts unicode.
+    """
     if PY3:
         string_type = str
     else:
@@ -704,13 +768,29 @@ class PtyProcessUnicode(PtyProcess):
         self.decoder = codecs.getincrementaldecoder(encoding)(errors=codec_errors)
 
     def read(self, size=1024):
+        """Read at most ``size`` bytes from the pty, return them as unicode.
+
+        Can block if there is nothing to read. Raises :exc:`EOFError` if the
+        terminal was closed.
+
+        The size argument still refers to bytes, not unicode code points.
+        """
         b = super(PtyProcessUnicode, self).read(size)
         return self.decoder.decode(b, final=False)
 
     def readline(self):
+        """Read one line from the pseudoterminal, and return it as unicode.
+
+        Can block if there is nothing to read. Raises :exc:`EOFError` if the
+        terminal was closed.
+        """
         b = super(PtyProcessUnicode, self).readline()
         return self.decoder.decode(b, final=False)
 
     def write(self, s):
+        """Write the unicode string ``s`` to the pseudoterminal.
+
+        Returns the number of bytes written.
+        """
         b = s.encode(self.encoding)
         return super(PtyProcessUnicode, self).write(b)
\ No newline at end of file
